from grapycal.sobjects.controls.control import Control
from objectsync import StringTopic, IntTopic


class TextControl(Control):
    '''
    To add a text control to a node, use the following code in the node:
    ```python
    self.add_control(TextControl, text='', label='', editable=True)
    ```
    '''
    frontend_type = 'TextControl'
    def build(self, text:str='', label:str='',readonly=False, editable:bool=True):
        if readonly:
            editable = False
        self.text = self.add_attribute('text', StringTopic, text, is_stateful= not readonly)
        self.label = self.add_attribute('label', StringTopic, label)
        self.editable = self.add_attribute('editable', IntTopic, 1 if editable else 0)
    
    def set(self, text:str):
        self.text.set(text)

    def get(self):
        return self.text.get()


import logging
import os

def setup_logging():
    logger = logging.getLogger()
    level = logging.WARNING
    #level = logging.DEBUG
    logging.basicConfig(level=level)
    ch = logging.StreamHandler()
    if os.environ.get('DEBUG') is not None and (os.environ.get('DEBUG').lower() == 'true'):
        ch.setLevel(logging.DEBUG)
    ch.setFormatter(ConsoleLogFormatter())
    if os.environ.get('DEBUG') is not None and (os.environ.get('DEBUG').lower() == 'true'):
        logger.setLevel(logging.DEBUG)
    logger.handlers.clear()
    logger.addHandler(ch)
    ch.addFilter(NameTranslator())

    return logger

class NameTranslator(logging.Filter):
    dictionary = {
        'objectsync.server': 'OS',
        'objectsync.sobject': 'SOBJECT',
        'chatroom.server.server': 'CHAT',
        'chatroom.topic': 'TOPIC',
        'chatroom.server.client_manager': 'CLIENT',
        
    }
    def filter(self, record):
        if record.name.startswith('websockets'):
            return False
        record.short_name = self.dictionary.get(record.name,record.name)
        return True

class ConsoleLogFormatter(logging.Formatter):

    green = "\x1b[32;20m"
    blue = "\x1b[34;20m"
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    bold = "\x1b[1m"
    no_bold = "\x1b[22m"
    reset = "\x1b[0m"
    #format = "%(threadName)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    format_ = f"%(levelname)5s {bold}[%(short_name)s]\t{no_bold} %(message)s"

    FORMATS = {
        'default':{
            logging.DEBUG: blue + format_ + reset,
            logging.INFO: grey + format_ + reset,
            logging.WARNING: yellow + format_ + reset,
            logging.ERROR: red + format_ + reset,
            logging.CRITICAL: bold_red + format_ + reset
        },
        'objectsync':{
            logging.DEBUG: green + format_ + reset,
            logging.INFO: grey + format_ + reset,
            logging.WARNING: yellow + format_ + reset,
            logging.ERROR: red + format_ + reset,
            logging.CRITICAL: bold_red + format_ + reset
        },
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.name.split('.')[0],self.FORMATS['default']).get(record.levelno)
        formatter = logging.Formatter(log_fmt)

        return formatter.format(record)


import random
from typing import Any
from grapycal.sobjects.port import InputPort, OutputPort, Port
from objectsync import SObject, StringTopic, ObjTopic, IntTopic
from objectsync.sobject import SObjectSerialized

try:
    import torch
    HAS_TORCH = True
except:
    HAS_TORCH = False

class Edge(SObject):
    frontend_type = 'Edge'

    def build(self, tail:OutputPort|None = None, head:InputPort|None = None):
        self.tail = self.add_attribute('tail', ObjTopic[OutputPort], tail)
        self.head = self.add_attribute('head', ObjTopic[InputPort], head)
        self.label = self.add_attribute('label', StringTopic, is_stateful=False)
        self.data_ready_topic = self.add_attribute('data_ready',IntTopic,1,is_stateful=False) # 0 for running, other for not running

    def init(self):                
        self._data = None
        self._activated = False
        self._data_ready = False
        self.reaquirable = False

        self.tail.on_set2 += self.on_tail_set
        self.head.on_set2 += self.on_head_set

        self.on_tail_set(None, self.tail.get())
        self.on_head_set(None, self.head.get())


    def on_tail_set(self, old_tail:Port|None, new_tail:Port|None):
        if old_tail:
            old_tail.remove_edge(self)
        if new_tail is None:
            self.remove()
            raise Exception(f'{self} tail cannot be None')
        if new_tail:
            new_tail.add_edge(self)
        self.label.set('')

    def on_head_set(self, old_head:Port|None, new_head:InputPort|None):
        if old_head:
            old_head.remove_edge(self)
        if new_head is None:
            self.remove()
            raise Exception(f'{self} head cannot be None')
        if new_head:
            new_head.add_edge(self)
            if self._activated:
                new_head.node.edge_activated(self, new_head)

    def destroy(self) -> SObjectSerialized:
        if self.tail.get():
            self.tail.get().remove_edge(self)
        if self.head.get():
            self.head.get().remove_edge(self)
        return super().destroy()

    def get_data(self)->Any:

        if not self._data_ready:
            raise Exception('Data not available')
        self._activated = False
        if not self.reaquirable:
            self._data_ready = False
            self.data_ready_topic.set(random.randint(0,10000))
        return self._data
    
    def push_data(self, data, label:str|None=None):
        self._data = data
        self._activated = True
        self._data_ready = True
        self.data_ready_topic.set(0)
        if label:
            self.label.set(label)
        else:
            label = ''
            if HAS_TORCH:
                if isinstance(data, torch.Tensor):
                    label = str(list(data.shape)) if list(data.shape)!=[] else 'scalar'
            self.label.set(label)

        head = self.head.get()
        if head:
            head.edge_activated(self)

        self._activated = False

    def set_label(self, label):
        self.label.set(label)
    
    def is_activated(self):
        return self._activated
    
    def is_data_ready(self):
        return self._data_ready



from typing import Any, Callable, List, TypeVar

T = TypeVar('T')
def as_type(x, t: type[T]) -> T:
    assert isinstance(x, t)
    return x


class Action:
    '''
    A hub for callbacks
    '''
    def __init__(self):
        self.callbacks:List[Callable] = []

    def __add__(self,callback:Callable):
        '''
        Temporary backward compatibility. To be removed.
        '''
        self.callbacks.append(callback)
        return self
    
    def __sub__(self,callback:Callable):
        '''
        Temporary backward compatibility. To be removed.
        '''
        self.callbacks.remove(callback)
        return self
    
    def invoke(self, *args: Any, **kwargs: Any) -> Any:
        '''Call each callback in the action with the given arguments.'''
        returns = []
        for callback in self.callbacks:
            returns.append(callback(*args,**kwargs))
        return returns


from grapycal.sobjects.edge import Edge
from grapycal.sobjects.port import InputPort
from grapycal.sobjects.node import Node
class SourceNode(Node):
    '''
    Inherit from this class to conveniently create a node that can be a source of the graph (can be activated by the user.)
    This type of node can be activated by double clicking the node or by sending a signal the run port on the node.
    '''
    def build_node(self):
        super().build_node()
        self.run_port = self.add_in_port('run')

    def init_node(self):
        super().init_node()
        self.run_port.on_activate += self.on_activate
    
    def task(self):
        '''
        Define the task of this node here. 
        By default, this method will be called when double clicking the node or when the run port on the node
          is activated (if there is one).
        '''
        pass

    def double_click(self):
        self.run(self.task)

    def on_activate(self, edge:Edge, port:InputPort):
        self.run_port.get_data() # clear data_ready so UI looks resonable 
        self.run(self.task)



import time
from grapycal.extension.extensionManager import ExtensionManager
from grapycal.extension.utils import Clock
from grapycal.sobjects.controls.threeControl import ThreeControl
from ..sobjects.controls import *
from grapycal.sobjects.editor import Editor
from grapycal.sobjects.workspaceObject import WebcamStream, WorkspaceObject
from grapycal.utils.io import file_exists, json_read, json_write
from grapycal.utils.logging import setup_logging
logger = setup_logging()

from typing import Any, Dict

import threading
import objectsync
from objectsync.sobject import SObjectSerialized
import asyncio
import signal
from dacite import from_dict


from . import stdout_helper


from grapycal.sobjects.edge import Edge
from grapycal.sobjects.port import InputPort, OutputPort
from grapycal.sobjects.sidebar import Sidebar

from grapycal.core.background_runner import BackgroundRunner
from grapycal.sobjects.node import Node

from . import running_module

def deserialize_sort_key(x: SObjectSerialized) -> int:
    if x.type == 'Edge':
        return 1000000000 # Edges should be created after nodes
    if x.id.startswith('0_'):
        return int(x.id[2:])
    return 0

class Workspace:
    def __init__(self, port, host, path) -> None:
        self.path = path
        self.running_module = running_module
        ''''''

        '''
        Enable stdout proxy for this process
        '''
        stdout_helper.enable_proxy(redirect_error=False)
        self.redirect = stdout_helper.redirect

        self._communication_event_loop: asyncio.AbstractEventLoop | None = None

        self.background_runner = BackgroundRunner()

        self._objectsync = objectsync.Server(port,host,deserialize_sort_key=deserialize_sort_key)
        
        self._extention_manager = ExtensionManager(self._objectsync,self)

        self.do_after_transition = self._objectsync.do_after_transition

        self.clock = Clock(0.1)

        self.webcam: WebcamStream|None = None

    def _communication_thread(self,event_loop_set_event: threading.Event):
        asyncio.run(self._async_communication_thread(event_loop_set_event))

    async def _async_communication_thread(self,event_loop_set_event: threading.Event):
        self._communication_event_loop = asyncio.get_event_loop()
        event_loop_set_event.set()
        await asyncio.gather(self._objectsync.serve(),self.clock.run())

    def run(self) -> None:
        print('Workspace running')
        event_loop_set_event = threading.Event()
        t = threading.Thread(target=self._communication_thread,daemon=True,args=[event_loop_set_event]) # daemon=True until we have a proper exit strategy

        t.start()
        event_loop_set_event.wait()

        self._objectsync.globals.workspace = self

        self._objectsync.register_service('exit', self.exit)

        self._objectsync.register(WorkspaceObject)
        self._objectsync.register(Editor)
        self._objectsync.register(Sidebar)
        self._objectsync.register(InputPort)
        self._objectsync.register(OutputPort)
        self._objectsync.register(Edge)

        self._objectsync.register(TextControl)
        self._objectsync.register(ButtonControl)
        self._objectsync.register(ImageControl)
        self._objectsync.register(ThreeControl)

        self._objectsync.register(WebcamStream)
        
        '''
        Register all built-in node types
        '''

        signal.signal(signal.SIGTERM, lambda sig, frame: self.exit())

        if file_exists(self.path):
            print(f'Found existing workspace file {self.path}. Loading')
            self.load_workspace(self.path)
        else:
            print(f'No workspace file found at {self.path}. Creating new workspace')
            self.initialize_workspace()
            
        self._objectsync.on('ctrl+s',lambda: self.save_workspace(self.path),is_stateful=False)
    
        self.background_runner.run()

    def exit(self):
        print('exit')
        self.background_runner.exit()

    def get_communication_event_loop(self) -> asyncio.AbstractEventLoop:
        assert self._communication_event_loop is not None
        return self._communication_event_loop
    
    '''
    Save and load
    '''         

    def initialize_workspace(self) -> None:
        self._objectsync.create_object(WorkspaceObject, parent_id='root')
        self._extention_manager.import_extension('grapycal_builtin')

    def save_workspace(self, path: str) -> None:
        workspace_serialized = self.get_workspace_object().serialize()
        data = {
            'extensions': self._extention_manager.get_extention_names(), 
            # Note: The extension field is intended to be on top so it can be easily retrieved by the extension management program.
            'client_id_count': self._objectsync.get_client_id_count(),
            'id_count': self._objectsync.get_id_count(),
            'workspace_serialized': workspace_serialized.to_dict(),
        }
        json_write(path, data)
        time_str = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        print(f'Workspace saved to {path} at {time_str}')

    def load_workspace(self, path: str) -> None:
        data = json_read(path)
        self._objectsync.set_client_id_count(data['client_id_count'])
        self._objectsync.set_id_count(data['id_count'])
        workspace_serialized = from_dict(SObjectSerialized,data['workspace_serialized'])
        self._extention_manager.load_extensions(data['extensions'])
        self._objectsync.create_object(WorkspaceObject, parent_id='root', serialized=workspace_serialized, id=workspace_serialized.id)

    def get_workspace_object(self) -> WorkspaceObject:
        # In case this called in self._objectsync.create_object(WorkspaceObject), 
        return self._objectsync.get_root_object().get_child_of_type(WorkspaceObject)
    
    def vars(self)->Dict[str,Any]:
        return self.running_module.__dict__

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--port', type=int, default=8765)
    parser.add_argument('--host', type=str, default='localhost')
    parser.add_argument('--path', type=str, default='workspace.grapycal')
    args = parser.parse_args()

    workspace = Workspace(args.port,args.host,args.path)
    workspace.run()


import logging

from grapycal.extension.utils import NodeInfo
logger = logging.getLogger(__name__)

from importlib import util as importlib_util
import random
from typing import TYPE_CHECKING, Dict, List, Tuple
import sys
from os.path import join, dirname
import shutil
from grapycal.extension.extension import Extension
from grapycal.sobjects.edge import Edge
from grapycal.sobjects.node import Node
from grapycal.sobjects.port import InputPort, OutputPort, Port
from grapycal.utils.file import get_direct_sub_folders
from grapycal.utils.misc import as_type
import objectsync

if TYPE_CHECKING:  
    from grapycal.core.workspace import Workspace

class ExtensionManager:
    def __init__(self,objectsync_server:objectsync.Server,workspace:'Workspace') -> None:
        self._objectsync = objectsync_server
        self._workspace = workspace
        cwd = sys.path[0]
        self._local_extension_dir = join(cwd,'.grapycal','extensions')
        sys.path.append(self._local_extension_dir)
        self._extensions: Dict[str, Extension] = {}

        # Use this topic to inform the client about the extensions
        self._imported_extensions_topic = self._objectsync.create_topic('imported_extensions',objectsync.DictTopic,is_stateful=False)
        self._avaliable_extensions_topic = self._objectsync.create_topic('avaliable_extensions',objectsync.DictTopic,is_stateful=False)
        self._node_types_topic = self._objectsync.create_topic('node_types',objectsync.DictTopic,is_stateful=False)
        self._objectsync.on('import_extension',self.import_extension,is_stateful=False)
        self._objectsync.on('unimport_extension',self.unimport_extension,is_stateful=False)
        self._objectsync.on('update_extension',self.update_extension,is_stateful=False)
        self._objectsync.on('refresh_extensions',self._update_available_extensions_topic,is_stateful=False)

        self._update_available_extensions_topic()

    def load_extensions(self,extension_names) -> None:
        for name in extension_names:
            if not Extension.extension_exists(name):
                logger.info(f'Extension {name} is not fetched. Fetching...')
                self._fetch_extension(name.split('_')[0]+'_'+name.split('_')[1],name.split('_')[2])
            self._load_extension(name)
        
        self._update_available_extensions_topic()

    def import_extension(self, package_name: str) -> Extension:
        name = self._fetch_extension(package_name)
        extension = self._load_extension(name)
        self._create_preview_nodes(name)
        self._update_available_extensions_topic()
        return extension

    def update_extension(self, extension_name: str) -> None:
        logger.info(f'Updating extension {extension_name}')
        old_version = self._extensions[extension_name]
        package_name = '_'.join(extension_name.split('_')[:-1])
        new_version = self._load_extension(self._fetch_extension(package_name))

        # Get diff between old and new version
        old_node_types = set(old_version.node_types_without_extension_name.keys())
        new_node_types = set(new_version.node_types_without_extension_name.keys())
        removed_node_types = old_node_types - new_node_types
        added_node_types = new_node_types - old_node_types
        changed_node_types = old_node_types & new_node_types

        existing_nodes = self._workspace.get_workspace_object().get_children_of_type(Node)

        # Ensure there are no nodes of removed types
        for node in existing_nodes:
            type_name_without_extension = node.get_type_name().split('.')[1]
            if type_name_without_extension in removed_node_types:
                self._unload_extension(extension_name)
                raise Exception(f'Cannot update extension {package_name} because there are still nodes of type {type_name_without_extension} in the workspace.')

        # Find nodes of changed types
        def hit(node:objectsync.SObject) -> bool:
            # Only update nodes that are not previews...
            if not isinstance(node,Node):
                return False
            if node.is_preview.get():
                return False
            
            # ... and are of the changed types.
            # Node type name format: grapycal_packagename_version.node_type_name
            node_type_name = node.get_type_name() 
            a = node_type_name.rfind('_',0,node_type_name.find('.'))
            b = node_type_name.find('.')
            node_package_name = node_type_name[:a]
            node_type_name_without_extension = node_type_name[b+1:]
            return node_package_name == package_name and node_type_name_without_extension in changed_node_types

        nodes_to_update = self._objectsync.get_root_object().top_down_search(
            accept=hit,
            stop=hit,
            type=Node
        )
        
        # Remove nodes of the changed types
        nodes_to_recover:List[Tuple[objectsync.sobject.SObjectSerialized,str]] = []
        wrapped_topics:set[Tuple[str,str]] = set()
        edges_to_recover:Dict[str,Tuple[str,str,str]] = {}
        removed_ports = set[str]()
        port_map_1 = {}
        for node in nodes_to_update:
            # First serialize the node
            nodes_to_recover.append((node.serialize(),node.get_parent().get_id()))

            for name, topic in node._attributes.items():
                if isinstance(topic,objectsync.WrappedTopic):
                    wrapped_topics.add((node.get_id(),name))

            # To handle edge recovery, we need to know the mapping between old and new port ids.
            # We use the port name and the node id to identify a port.
            # The mapping can be accessed by combining the following maps:
            #   port_map_1: old id -> (port name, old node id)
            #   node_id_map: old node id -> new node id
            #   port_map_2: (port name, new node id) -> new id
            ports: List[Port] = node.in_ports.get() + node.out_ports.get() # type: ignore
            for port in ports:
                port_map_1[port.get_id()] = (port.is_input.get(),port.name.get(),node.get_id())
            for port in ports:
                removed_ports.add(port.get_id())
                for edge in port.edges.copy():
                    edges_to_recover[edge.get_id()] = (edge.tail._topic.get(),edge.head._topic.get(),edge.get_parent().get_id())
                    self._objectsync.destroy_object(edge.get_id())

            # Then destroy the node
            self._objectsync.destroy_object(node.get_id())


        '''
        Now, the old nodes, ports and edges are destroyed. Their information is stored in nodes_to_recover and edges_to_recover.
        '''
        
        # Recover nodes with new version
        
        node_id_map = {}
        new_nodes:Dict[str,Tuple[objectsync.sobject.SObjectSerialized,Node]] = {}
        for old_serialized, parent_id in nodes_to_recover:
            type_name = f'{new_version.extension_name}.{old_serialized.type.split(".")[1]}'

            # Create the new node instance
            new_node: Node = self._objectsync.create_object_s(type_name,parent_id=parent_id) # type: ignore
            new_nodes[new_node.get_id()] = (old_serialized,new_node)
            node_id_map[old_serialized.id] = new_node.get_id()
        
        # Update values of wrapped topics
        for old_serialized,new_node in new_nodes.values():
            for attr in old_serialized.attributes:
                if (old_serialized.id,attr[0]) in wrapped_topics:
                    if isinstance(attr[2],list):
                        tmp = []
                        for value in attr[2]:
                            if value in node_id_map:
                                tmp.append(node_id_map[value])
                            else:
                                tmp.append(value)
                        attr[2] = tmp
                    elif isinstance(attr[2],dict):
                        tmp = {}
                        for key,value in attr[2].items():
                            if value in node_id_map:
                                tmp[key] = node_id_map[value]
                            else:
                                tmp[key] = value
                        attr[2] = tmp
                    else:
                        attr[2] = node_id_map[attr[2]] if attr[2] in node_id_map else attr[2]
                        

        # Recover the new nodes
        port_map_2 = {}
        for old_serialized,new_node in new_nodes.values():
            # Let the node handle the recovery
            old_node_info = NodeInfo(old_serialized)
            new_node.old_node_info = old_node_info
            new_node.restore_from_version('',old_node_info) #TODO version name


            ports: List[Port] = new_node.in_ports.get() + new_node.out_ports.get() # type: ignore
            for port in ports:
                port_map_2[(port.is_input.get(),port.name.get(),new_node.get_id())] = port.get_id()
            
            #TODO: let the new node class handle the recovery for backwards compatibility

        logger.info(f'Updated extension {package_name} from {old_version.extension_name} to {new_version.extension_name}')


        # Recover edges if possible
        def port_id_map(old_port_id:str) -> str|None:
            
            try:
                is_input, port_name, old_node_id = port_map_1[old_port_id]
            except KeyError:
                return None
            
            new_node_id = node_id_map[old_node_id]

            try:
                new_port_id =  port_map_2[(is_input,port_name,new_node_id)]
            except KeyError:
                return None
            
            return new_port_id

        for tail_id, head_id, parent_id in edges_to_recover.values():
            #print('tail_id:',tail_id,'head_id:',head_id,'parent_id:',parent_id,port_map_1,port_map_2,node_id_map)
            if tail_id not in removed_ports:
                new_tail_id = tail_id
            else:
                new_tail_id = port_id_map(tail_id)
            if head_id not in removed_ports:
                new_head_id = head_id
            else:
                new_head_id = port_id_map(head_id)

            if new_tail_id is None or new_head_id is None:
                continue # The port does not present in the new version, unfortunately we cannot recover the edge.
            
            new_tail = self._objectsync.get_object(new_tail_id)
            new_head = self._objectsync.get_object(new_head_id)
            new_edge = self._objectsync.create_object(Edge,parent_id=parent_id,tail=new_tail,head=new_head)

        # Unimport old version
        self.unimport_extension(old_version.extension_name)
        self._create_preview_nodes(new_version.extension_name)

        #TODO: cut history

    def unimport_extension(self, extension_name: str) -> None:
        self._check_extension_not_used(extension_name)
        self._destroy_preview_nodes(extension_name)
        self._unload_extension(extension_name)
        self._update_available_extensions_topic()
        # Don't unfetch now, because the user might change their mind.
        # The Grapycal App will unfetch the unused extension when closing.

    def _update_available_extensions_topic(self) -> None:
        available_extensions = self._scan_available_extensions()
        for name in available_extensions:
            if name not in self._avaliable_extensions_topic.get():
                self._avaliable_extensions_topic.add(name,{
                    'name':name
                })
        for name in list(self._avaliable_extensions_topic.get().keys()):
            if name not in available_extensions:
                self._avaliable_extensions_topic.pop(name)

    def _scan_available_extensions(self) -> list[str]:
        '''
        Returns a list of available extensions in local folder.
        '''
        available_extensions = []
        # Local
        for name in get_direct_sub_folders('.'):
            if not name.startswith('grapycal_'):
                continue
            continue_flag = False
            for extension in self._extensions.values():
                if name == extension.package_name:
                    continue_flag = True
                    break
            if continue_flag:
                continue

            available_extensions.append(name)

        # TODO: pip
        return available_extensions

    def _fetch_extension(self, package_name: str, number=None) -> str:
        '''
        Copy the current version of the extension to .grapycal/extensions, so it cannot be modified by the user.
        '''
        assert package_name.startswith('grapycal_'), f'Extension name must start with grapycal_, got {package_name}'
        source_name = package_name
        source_path = dirname(importlib_util.find_spec(source_name).origin) # type: ignore

        if number is None:
            number = random.randint(0,1000000)
        extension_name = f'{package_name}_{number}'
        assert source_path is not None
        shutil.copytree(source_path,join(self._local_extension_dir,extension_name),dirs_exist_ok=True)
        return extension_name

    def _load_extension(self, name: str) -> Extension:
        self._extensions[name] = Extension(name,self._objectsync.get_all_node_types())
        for node_type_name, node_type in self._extensions[name].node_types.items():
            self._objectsync.register(node_type,node_type_name)
        self._imported_extensions_topic.add(name,{
            'name':name
        })
        for node_type_name, node_type in self._extensions[name].node_types.items():
            self._node_types_topic.add(node_type_name,{
                'name':node_type_name,
                'category':node_type.category
            })
        return self._extensions[name]
    
    def _check_extension_not_used(self, name: str) -> None:
        node_types = self._extensions[name].node_types
        for obj in self._objectsync.get_objects():
            if isinstance(obj, Node):
                if not obj.is_preview.get() and obj.get_type_name() in node_types:
                    raise Exception(f'Cannot unload extension {name}, there are still objects of this type in the workspace')

    def _unload_extension(self, name: str) -> None:
        node_types = self._extensions[name].node_types
        for node_type in node_types:
            self._objectsync.unregister(node_type)
        self._extensions.pop(name)
        self._imported_extensions_topic.pop(name)
        for node_type_name in node_types:
            self._node_types_topic.pop(node_type_name)
    
    def _create_preview_nodes(self, name: str) -> None:
        node_types = self._extensions[name].node_types
        for node_type in node_types.values():
            if not node_type.category == 'hidden':
                self._objectsync.create_object(node_type,parent_id=self._workspace.get_workspace_object().sidebar.get_id(),is_preview=True)
 
    def _destroy_preview_nodes(self, name: str) -> None:
        node_types = self._extensions[name].node_types
        for obj in self._workspace.get_workspace_object().sidebar.get_children_of_type(Node):
            if obj.get_type_name() in node_types:
                self._objectsync.destroy_object(obj.get_id())

    def get_extension(self, name: str) -> Extension:
        return self._extensions[name]
    
    def get_extention_names(self) -> list[str]:
        return list(self._extensions.keys())


'''
https://stackoverflow.com/questions/14890997/redirect-stdout-to-a-file-only-for-a-specific-thread#:~:text=You%20cannot%20redirect%20stdout%20for,messes%20with%20all%20of%20them.
'''

from contextlib import contextmanager
import threading
import sys
import io
from typing import Any
from . import local

# Save all of the objects for use later.
orig___stdout__ = sys.__stdout__
orig___stderr__ = sys.__stderr__
orig_stdout = sys.stdout
orig_stderr = sys.stderr
thread_proxies = {}

def orig_print(*args,flush=False):
    orig___stdout__.write(' '.join([str(arg) for arg in args])+'\n')

# def redirect():
#     """
#     Enables the redirect for the current thread's output to a single io
#     object and returns the object.

#     :return: The StringIO object.
#     :rtype: ``io.StringIO``
#     """
#     # Get the current thread's identity.
#     ident = threading.currentThread().ident

#     # Enable the redirect and return the io object.
#     thread_proxies[ident] = io.StringIO()
#     return thread_proxies[ident]


# def stop_redirect():
#     """
#     Enables the redirect for the current thread's output to a single io
#     object and returns the object.

#     :return: The final string value.
#     :rtype: ``str``
#     """
#     # Get the current thread's identity.
#     ident = threading.currentThread().ident

#     # Only act on proxied threads.
#     if ident not in thread_proxies:
#         return

#     # Read the value, close/remove the buffer, and return the value.
#     retval = thread_proxies[ident].getvalue()
#     thread_proxies[ident].close()
#     del thread_proxies[ident]
#     return retval

@contextmanager
def redirect(stringio: Any):
    """
    Context manager for redirecting stdout to a single io object.
    """
    ident = threading.currentThread().ident
    old_proxy = thread_proxies.get(ident, None)
    thread_proxies[ident] = stringio
    try:
        yield
    finally:
        # possible reentrant call
        if old_proxy is None:
            del thread_proxies[ident]
        else:
            thread_proxies[ident] = old_proxy

def _get_stream(original):
    """
    Returns the inner function for use in the LocalProxy object.

    :param original: The stream to be returned if thread is not proxied.
    :type original: ``file``
    :return: The inner function for use in the LocalProxy object.
    :rtype: ``function``
    """
    def proxy():
        """
        Returns the original stream if the current thread is not proxied,
        otherwise we return the proxied item.

        :return: The stream object for the current thread.
        :rtype: ``file``
        """
        # Get the current thread's identity.
        ident = threading.currentThread().ident

        # Return the proxy, otherwise return the original.
        return thread_proxies.get(ident, original)

    # Return the inner function.
    return proxy


def enable_proxy(redirect_error=True):
    """
    Overwrites __stdout__, __stderr__, stdout, and stderr with the proxied
    objects.
    """
    sys.__stdout__ = local.LocalProxy(_get_stream(sys.__stdout__))
    sys.stdout = local.LocalProxy(_get_stream(sys.stdout))

    if redirect_error:
        sys.__stderr__ = local.LocalProxy(_get_stream(sys.__stderr__))
        sys.stderr = local.LocalProxy(_get_stream(sys.stderr))


def disable_proxy():
    """
    Overwrites __stdout__, __stderr__, stdout, and stderr with the original
    objects.
    """
    sys.__stdout__ = orig___stdout__
    sys.__stderr__ = orig___stderr__
    sys.stdout = orig_stdout
    sys.stderr = orig_stderr


import logging
logger = logging.getLogger(__name__)

from collections import deque
from contextlib import contextmanager
from queue import Queue
import queue
import traceback
from typing import Callable
import signal
from .stdout_helper import orig_print

class BackgroundRunner:
    def __init__(self):
        self._inputs: Queue = Queue()
        self._queue: deque = deque()
        self._stack: deque = deque()
        self._exit_flag = False
        self._exception_callback: Callable[[Exception], None] = lambda e: orig_print('runner default exception callback:\n',traceback.format_exc())

    def push(self, task: Callable, to_queue: bool = True):
        self._inputs.put((task, to_queue))

    def push_to_queue(self, task: Callable):
        self._inputs.put((task, True))

    def push_to_stack(self, task: Callable):
        self._inputs.put((task, False))

    def interrupt(self):
        signal.raise_signal(signal.SIGINT)

    def clear_tasks(self):
        self._queue.clear()
        self._stack.clear()

    def set_exception_callback(self, callback: Callable[[Exception], None]):
        self._exception_callback = callback

    def exit(self):
        self._exit_flag = True
        self.interrupt()

    @contextmanager
    def no_interrupt(self):
        def handler(signum, frame):
            logger.info("no_interrupt: continue")

        original_sigint_handler = signal.getsignal(signal.SIGINT)
        try:
            signal.signal(signal.SIGINT, handler)
            yield
        finally:
            signal.signal(signal.SIGINT, original_sigint_handler)

    def run(self):
        while True:
            if self._exit_flag:
                break
            try:
                # Queue.get() blocks signal.
                while not self._inputs.empty() or (len(self._queue) == 0 and len(self._stack) == 0):
                    try:
                        inp = self._inputs.get(timeout=0.2)
                    except queue.Empty:
                        continue
                    task, push_to_queue = inp
                    if push_to_queue:
                        self._queue.appendleft(task)
                    else:
                        self._stack.append(task)

                logger.debug('got a task OAO')

                # queue is prioritized
                if len(self._queue) > 0:
                    task_to_run = self._queue.pop()
                else:
                    task_to_run = self._stack.pop()

                task_to_run()

            except KeyboardInterrupt:
                logger.info("runner catch keyboardinterrupt")

            except Exception as e:
                self.clear_tasks()
                self._exception_callback(e)



from typing import Any, Callable, Dict, Generic, List, TypeVar
from grapycal.utils.misc import Action
from objectsync.sobject import SObjectSerialized
import asyncio

I=TypeVar('I')
O=TypeVar('O')
class LazyDict(Generic[I,O]):
    def __init__(self,gen:Callable[[I],O],keys:List[I]) -> None:
        self.gen = gen
        self._keys=keys

    def __getitem__(self,idx:I)->O:
        return self.gen(idx)
    
    def __contains__(self,idx:I)->bool:
        return idx in self._keys
    
    def keys(self):
        return self._keys


class AttrInfo:
    def __init__(self, name, type_name, value, is_stateful):
        self.name = name
        self.type = type_name
        self.value = value
        self.is_stateful = is_stateful

class SObjectInfo:
    def __init__(self,serialized:SObjectSerialized):
        self.serialized = serialized
        self.attr_info: Dict[str,AttrInfo] = {}
        self.attributes: Dict[str,Any] = {}
        for name, type_name, value, is_stateful in self.serialized.attributes:
            self.attr_info[name] = AttrInfo(name, type_name, value, is_stateful)
            self.attributes[name] = value

        

    def has_attribute(self,name):
        return name in self.attr_info
    
    def __getitem__(self,name:str):
        '''
        Returns the value of an attribute
        '''
        return self.attributes[name]

class ControlInfo(SObjectInfo):
    pass

class NodeInfo(SObjectInfo):
    '''
    An easier-to-use interface to read SObjectSerialized of a node
    '''
    def __init__(self, serialized: SObjectSerialized):
        super().__init__(serialized)
        children = serialized.children
        controls: Dict[str,str] = self.attributes['controls']
        self.controls=LazyDict[str,ControlInfo](lambda name: ControlInfo(children[controls[name]]),list(controls.keys()))

class Clock:
    def __init__(self, interval: float):
        self.interval = interval
        self.on_tick = Action()


    async def run(self):
        while True:
            await asyncio.sleep(self.interval)
            self.on_tick.invoke()


import io
from typing import Any, Dict
from grapycal.extension.utils import ControlInfo
from grapycal.sobjects.controls.control import Control
from objectsync import StringTopic

smallest_jpg = '/9j/4AAQSkZJRgABAQEASABIAAD/2wBDAAMCAgICAgMCAgIDAwMDBAYEBAQEBAgGBgUGCQgKCgkICQkKDA8MCgsOCwkJDRENDg8QEBEQCgwSExIQEw8QEBD/yQALCAABAAEBAREA/8wABgAQEAX/2gAIAQEAAD8A0s8g/9k='

class ImageControl(Control):
    '''
    '''
    frontend_type = 'ImageControl'
    def build(self):
        self.image = self.add_attribute('image', StringTopic,smallest_jpg,is_stateful=False)
        self.on_image_set = self.image.on_set

    def restore_from(self, old: ControlInfo):
        super().restore_from(old)
        self.image.set(old['image'])
    
    def set(self,image:bytes|io.BytesIO|None):
        if image is None:
            self.image.set(smallest_jpg)
            return
        if isinstance(image,io.BytesIO):
            image.seek(0)
            image = image.read()
        self.image.set_from_binary(image)

    def get(self) -> bytes:
        return self.image.to_binary()
        


from typing import TYPE_CHECKING, Any, List
from objectsync import SObject, StringTopic, IntTopic
from grapycal.utils.misc import Action

if TYPE_CHECKING:
    from grapycal.sobjects.node import Node
    from grapycal.sobjects.edge import Edge

class Port(SObject):
    frontend_type = 'Port'

    def build(self, name='port', max_edges=64, display_name=None):
        self.name = self.add_attribute('name', StringTopic, name)
        self.display_name = self.add_attribute('display_name', StringTopic, name if display_name is None else display_name)
        self.max_edges = self.add_attribute('max_edges', IntTopic, max_edges)
        self.is_input = self.add_attribute('is_input', IntTopic, 0)

    def init(self):
        self.edges: List[Edge] = []
        self.node: Node = self.get_parent() # type: ignore
    
    def add_edge(self, edge:'Edge'):
        if len(self.edges) >= self.max_edges.get():
            raise Exception('Max edges reached')
        self.edges.append(edge)
    
    def remove_edge(self, edge:'Edge'):
        if edge not in self.edges:
            return
        self.edges.remove(edge)

    def is_full(self):
        return len(self.edges) >= self.max_edges.get()

    def get_name(self):
        return self.name.get()

class InputPort(Port):
    def build(self, name='port', max_edges=64, display_name=None):
        super().build(name, max_edges, display_name)
        self.is_input.set(1)

    def init(self):
        super().init()
        self.on_activate = Action()

    def add_edge(self, edge:'Edge'):
        super().add_edge(edge)
        self.node.input_edge_added(edge, self)

    def remove_edge(self, edge:'Edge'):
        super().remove_edge(edge)
        self.node.input_edge_removed(edge, self)
        
    def is_all_edge_ready(self):
        return all(edge.is_data_ready() for edge in self.edges) and len(self.edges) > 0
    
    def get_data(self):
        return [edge.get_data() for edge in self.edges]
    
    def get_one_data(self,allow_no_data=False):
        if allow_no_data and not self.is_all_edge_ready():
            return None
        return self.edges[0].get_data()
    
    def edge_activated(self, edge:'Edge'):
        self.on_activate.invoke(self, edge)
        self.node.edge_activated(edge, self)

class OutputPort(Port):
    def build(self, name='port', max_edges=64, display_name=None):
        super().build(name, max_edges, display_name)
        self.is_input.set(0)

    def init(self):
        super().init()
        self._retain = False
        self._retained_data = None

    def add_edge(self, edge:'Edge'):
        super().add_edge(edge)
        if self._retain:
            edge.push_data(self._retained_data)
        self.node.output_edge_added(edge, self)

    def remove_edge(self, edge:'Edge'):
        super().remove_edge(edge)
        self.node.output_edge_removed(edge, self)

    def push_data(self, data:Any=None,label:str|None=None, retain: bool = False):
        '''
        Push data to all connected edges.
        If retain is True, the data will be pushed to all future edges when they're connected as well.
        '''
        if retain:
            self._retain = True
            self._retained_data = data
        for edge in self.edges:
            edge.push_data(data,label=label)

    def disable_retain(self):
        '''
        Disable retain mode.
        '''
        self._retain = False
        self._retained_data = None # Release memory



import importlib
import inspect
from typing import Dict
from warnings import warn

from grapycal.sobjects.node import Node
from objectsync import SObject

class Extension:
    @staticmethod
    def extension_exists(extension_name:str) -> bool:
        try:
            importlib.import_module(extension_name)
            return True
        except ModuleNotFoundError:
            return False
    def __init__(self,extension_name:str,existing_node_types:Dict[str,type[SObject]]={}) -> None:
        self.module = importlib.import_module(extension_name)
        self.extension_name = extension_name
        self.package_name ='_'.join( extension_name.split('_')[:-1])

        self.node_types:Dict[str,type[Node]] = {}
        self.node_types_without_extension_name:Dict[str,type[Node]] = {}
        for name, obj in inspect.getmembers(self.module):
            if inspect.isclass(obj) and issubclass(obj, Node):
                type_name = f'{self.extension_name}.{obj.__name__}'

                '''
                It may happen that the same node type is defined in multiple extensions. It is possible caused
                by diamond-imports. For example, grapycal_ext1 and grapycal_ext2 both import FunctionNode from 
                grapycal.sobjects.functionNode, and both mistakenly expose it in their __init__.py. This will
                cause an error when trying to register the node type to ObjectSync. To prevent this, we check
                if the node type already exists in the existing_node_types dict. If it does, we skip it.
                Not very elegant, but hope it works.
                '''
                if obj in existing_node_types.values():
                    warn(f'Node type {type_name} already exists. Skipping')
                    continue

                self.node_types[type_name] = obj
                self.node_types_without_extension_name[obj.__name__] = obj



from typing import Any
from grapycal.sobjects.edge import Edge
from grapycal.sobjects.node import Node
class FunctionNode(Node):
    inputs = []
    max_in_degree = []
    outputs = []
    display_port_names = True

    def build_node(self):
        self._max_in_degree = self.max_in_degree[:]
        while len(self._max_in_degree) < len(self.inputs):
            self._max_in_degree.append(None)
        for i in range(len(self._max_in_degree)):
            if self._max_in_degree[i] is None:
                self._max_in_degree[i] = 1024
        for name, max_edges in zip(self.inputs,self._max_in_degree): #type: ignore
            display_name = name if self.display_port_names else ''
            self.add_in_port(name,max_edges,display_name=display_name)
        for name in self.outputs:
            display_name = name if self.display_port_names else ''
            self.add_out_port(name,display_name=display_name)

        self.label.set('f')
        self.shape.set('round')

    def edge_activated(self, edge: Edge, port):
        for port in self.in_ports:
            if not port.is_all_edge_ready():
                return
        self.run(self.task)

    def task(self):
        if self.is_destroyed():
            return
        inputs = {}
        for port in self.in_ports:
            if port.max_edges.get() == 1:
                inputs[port.get_name()] = port.edges[0].get_data()
            else:
                inputs[port.get_name()] = [edge.get_data() for edge in port.edges]
        result = self.calculate(**inputs)

        if result is None:
            return

        if len(self.out_ports) == 1:
            self.out_ports[0].push_data(result)
        else:
            for k,v in result.items():
                self.get_out_port(k).push_data(v)

    def calculate(self, **inputs)->Any:
        '''
        Define the function of this node here.

        :param **inputs: A dict of lists. Each dict entry is the data from one input port. For example, if there is only one input port named "in" with two\
            edges connected, the value of inputs will be `{'in':[data_from_edge_1, data_from_edge_2]}`.
        :return: A list of outputs. Each entry in the list will be pushed to one output port. For example, if there is only one output port,\
            the value of the return suold be `[data_to_port]`. If there are multiple edges connected to one output port,\
            they will all receive the same data.

        Examples::

            class SumNode(FunctionNode):
                inputs = ['items']
                outputs = ['sum']
                ...
                def calculate(self, items):
                    return {'sum':sum(items)} # Return the sum of the data
        '''
        raise NotImplementedError

    def input_edge_added(self, edge: Edge, port):
        for port in self.in_ports:
            if not port.is_all_edge_ready():
                return
        self.run(self.task)

    def input_edge_removed(self, edge: Edge, port):
        for port in self.in_ports:
            if not port.is_all_edge_ready():
                return
        self.run(self.task)

    def remove(self):
        return super().remove()


from grapycal.sobjects.edge import Edge
from grapycal.sobjects.node import Node
from grapycal.sobjects.port import InputPort, OutputPort
from objectsync import SObject

class Editor(SObject):
    frontend_type = 'Editor'

    def init(self):
        # called by client
        self.register_service('create_edge',self.create_edge_service)
        self.register_service('create_node',self.create_node_service)

    def create_node(self, node_type: type, **kwargs) -> Node:
        return self.add_child(node_type, is_preview=False, **kwargs)
    
    def create_node_service(self, node_type: str, **kwargs):
        self.add_child_s(node_type, is_preview=False, **kwargs)
    
    def create_edge(self, tail: OutputPort, head: InputPort) -> Edge:
        # Check the tail and head have space for the edge
        if tail.is_full() or head.is_full():
            raise Exception('A port is full')
        return self.add_child(Edge, tail=tail, head=head)
    
    def create_edge_service(self, tail_id: str, head_id: str):
        tail = self._server.get_object(tail_id)
        head = self._server.get_object(head_id)
        assert isinstance(tail, OutputPort)
        assert isinstance(head, InputPort)
        self.create_edge(tail, head)



from grapycal.sobjects.controls.control import Control
from objectsync import StringTopic, IntTopic, ListTopic


class ThreeControl(Control):
    '''
    '''
    frontend_type = 'ThreeControl'
    def build(self):
        super().build()
        self.points = self.add_attribute('points',ListTopic)
        self.lines = self.add_attribute('lines',ListTopic)


import logging
logger = logging.getLogger(__name__)

import json
import asyncio
import io
import threading
from typing import Any, Callable

class OutputStream:
    def __init__(self, on_flush:Callable[[str],None], hz=20):
        self._stream = io.StringIO()
        self._enables = 0
        self._enable_flush_event = asyncio.Event()
        self._lock = threading.Lock()
        self._exit_flag = False
        self._on_flush = on_flush
        self._gap = 1/hz

    def set_event_loop(self, event_loop):
        self._event_loop = event_loop

    async def run(self):
        self._event_loop = asyncio.get_running_loop()
        while True:
            if self._exit_flag:
                self._stream.close()
                return
            await self._enable_flush_event.wait()
            if self._exit_flag:
                self._stream.close()
                return
            while self._enables > 0:
                if self._exit_flag:
                    self._stream.close()
                    return
                # read
                self._stream.seek(0)
                data = self._stream.read()
                # clear
                self._stream.seek(0)
                self._stream.truncate()
                self._on_flush(data)
                await asyncio.sleep(self._gap)

            self._enable_flush_event.clear()

            # read
            self._stream.seek(0)
            data = self._stream.read()
            # clear
            self._stream.seek(0)
            self._stream.truncate()
            self._on_flush(data)

    def write(self, data):
        # Is stream thread safe?
        # locking makes deadlock
        self._stream.write(data)

    def flush(self): # dummy
        return
    
    def enable_flush(self):
        self._event_loop.call_soon_threadsafe(self._enable_flush)
        
    def _enable_flush(self):
        self._enables += 1
        self._enable_flush_event.set()

    def disable_flush(self):
        self._event_loop.call_soon_threadsafe(self._disable_flush)

    def _disable_flush(self):
        self._enables -= 1

    def close(self):
        self._exit_flag = True
        self._enable_flush_event.set()

def json_write(path:str,data:Any):
    with open(path,'w') as f:
        json.dump(data,f)

def json_read(path):
    with open(path,'r') as f:
        return json.load(f)
    
def file_exists(path):
    try:
        with open(path,'r') as f:
            return True
    except FileNotFoundError:
        return False


from grapycal.app import GrapycalApp
from grapycal.sobjects.sourceNode import SourceNode
from grapycal.utils.config import load_config
from grapycal.sobjects.controls import TextControl, ButtonControl, ImageControl
from grapycal.sobjects.node import Node
from grapycal.sobjects.functionNode import FunctionNode
from grapycal.sobjects.edge import Edge
from grapycal.sobjects.port import InputPort, OutputPort, Port
from objectsync import Topic, IntTopic, SetTopic, DictTopic, StringTopic, ListTopic, GenericTopic, FloatTopic, EventTopic, ObjListTopic, ObjSetTopic, ObjDictTopic, ObjTopic


from grapycal.sobjects.edge import Edge
from grapycal.sobjects.editor import Editor
from grapycal.sobjects.node import Node
from grapycal.sobjects.sidebar import Sidebar
from objectsync import SObject, ObjTopic, StringTopic, IntTopic

class WorkspaceObject(SObject):
    frontend_type = 'Workspace'
        
    def build(self):
        self.webcam = self.add_child(WebcamStream)
        self.sidebar =self.add_child(Sidebar)
        self.main_editor = self.add_child(Editor)
        self.add_attribute('main_editor',ObjTopic).set(self.main_editor)

    def init(self):
        self._server.on('delete',self._delete_callback,is_stateful=False)

    def _delete_callback(self,ids):
        nodes:set[Node] = set()
        edges:set[Edge] = set()
        for id in ids:
            obj = self._server.get_object(id)
            match obj:
                case Node():
                    nodes.add(obj)
                case Edge():
                    edges.add(obj)

        for node in nodes:
            for port in node.in_ports.get() + node.out_ports.get():
                for edge in port.edges:
                    edges.add(edge)
                    
        for edge in edges:
            if edge.is_destroyed():
                raise Exception(f'Edge {edge} is already destroyed')
        for node in nodes:
            if node.is_destroyed():
                raise Exception(f'Node {node} is already destroyed')

        for edge in edges:
            edge.remove()
        for node in nodes:
            node.remove()

class WebcamStream(SObject):
    frontend_type = 'WebcamStream'
    def build(self):
        self.image = self.add_attribute('image',StringTopic)
        self.source_client = self.add_attribute('source_client',IntTopic,-1)

    def init(self):
        self.source_client.set(-1)
        self._server.globals.workspace.webcam = self


import os
import shutil
import signal
import sys
from typing import Any, Callable, Dict
import subprocess
from .utils.file import get_direct_sub_folders
import termcolor
import time
from importlib.metadata import version as get_version

class GrapycalApp:
    """
    The backend server

    :param usersettings.Settings config: the configuration for server
    """
    def __init__(self, config) -> None:
        self._config = config

    def run(self) -> None:
        """
        Server main loop
        """
        version = get_version('grapycal')
        print(
            termcolor.colored(r'''
               ______                                  __
              / ____/________ _____  __  ___________ _/ /
             / / __/ ___/ __ `/ __ \/ / / / ___/ __ `/ / 
            / /_/ / /  / /_/ / /_/ / /_/ / /__/ /_/ / /  
            \____/_/   \__,_/ .___/\__, /\___/\__,_/_/   
                           /_/    /____/
                                           ''','red') + termcolor.colored('v'+version, 'grey'))
        print()
        print('Starting Grapycal server...')

        if not os.path.exists('./.grapycal'):
            os.mkdir('./.grapycal')
        if not os.path.exists('./.grapycal/extensions'):
            os.mkdir('./.grapycal/extensions')
            
        self.clean_unused_fetched_extensions()

        #TODO: Support multiple workspaces
        #TODO: Websocket multiplexing

        workspace = None
        # Here simply start one workspace in background
        while True:
            break_flag = False
            try:
            
                # Start webpage server
                if not self._config['no_serve_webpage']:
                    webpage_path = os.path.join(os.path.dirname(__file__),'webpage')
                    print(f'Strating webpage server at localhost:9001 from {webpage_path}')
                    subprocess.Popen([sys.executable,'-m', 'http.server','9001'],start_new_session=True,cwd=webpage_path)

                while True: # Restart workspace when it exits. Convenient for development

                    # Start workspace
                    print(f'Starting workspace {self._config["path"]}')
                    print(f'Host: {self._config["host"]}')
                    print(f'Port: {self._config["port"]}')
                    workspace = subprocess.Popen([sys.executable,'-m', 'grapycal.core.workspace',
                        '--port', str(self._config['port']),
                        '--host', self._config['host'],
                        '--path', self._config['path']],start_new_session=True)
                    
                    while True:
                        time.sleep(1)
                        if workspace.poll() is not None:
                            print('Workspace exited with code',workspace.poll())
                            break

                    if not self._config['restart']:
                        break

            except KeyboardInterrupt:
                print('Shutting down Grapycal server will force kill all workspaces. Press Ctrl+C again to confirm.')
                try:
                    time.sleep(3)
                except KeyboardInterrupt:
                    break_flag = True
                else:
                    print('Resumed')
            if break_flag:
                break

        print('Stopping workspaces...')
        if workspace:
            workspace.send_signal(signal.SIGTERM)
            workspace.wait()
        self.clean_unused_fetched_extensions()
        print('Grapycal server terminated')
        return
    
    def clean_unused_fetched_extensions(self):
        used_extensions = []
        for f in os.listdir('.'):
            if f.endswith('.grapycal') and not f.startswith('.'):
                used_extensions += self.parse_extensions_from_workspace(f)
        for dir in get_direct_sub_folders('./.grapycal/extensions'):
            if dir not in used_extensions:
                shutil.rmtree(f'./.grapycal/extensions/{dir}')
                
    def parse_extensions_from_workspace(self,workspace_path: str) -> list[str]:
        extensions_field = ''
        pre_match = '{"extensions":['
        with open(workspace_path,'r') as f:
            while len(pre_match) > 0:
                char = f.read(1)
                if char == ' ' or char == '\n':
                    continue
                assert char == pre_match[0], f'Expected {pre_match[0]} but got {char}'
                pre_match = pre_match[1:]
            while True:
                char = f.read(1)
                if char == ' ' or char == '\n' or char == '"':
                    continue
                if char == ']':
                    break
                extensions_field += char
        return extensions_field.split(',')





from grapycal.extension.utils import ControlInfo
from objectsync import SObject

class Control(SObject):
    '''
    A control enables the user to interact with a node via the GUI. There are different types of controls, such as a text input control
    a slider control, a checkbox control, etc. Controls can be added to a node during the Node.build() process or dynamically added or removed at runtime.
    '''
    def restore_from(self,old:ControlInfo):
        '''
        Default recovery process. If the control class get updated in Grapycal, override this method to customize the recovery process
        for different Grapycal versions.
        '''
        for k,v in old.attributes.items():
            self.get_attribute(k).set(v)
        
         


from abc import ABCMeta
import io
from itertools import count
import logging
import random
from grapycal.sobjects.controls.buttonControl import ButtonControl
from grapycal.sobjects.controls.imageControl import ImageControl

from grapycal.sobjects.controls.textControl import TextControl
logger = logging.getLogger(__name__)
from contextlib import contextmanager
import functools
import traceback
from typing import TYPE_CHECKING, Any, Callable, TypeVar
from grapycal.extension.utils import NodeInfo
from grapycal.sobjects.controls.control import Control
from grapycal.sobjects.edge import Edge
from grapycal.sobjects.port import InputPort, OutputPort
from grapycal.utils.io import OutputStream
from objectsync import SObject, StringTopic, IntTopic, ListTopic, ObjListTopic, FloatTopic, Topic, ObjDictTopic, SetTopic
from objectsync.sobject import SObjectSerialized, WrappedTopic

if TYPE_CHECKING:
    from grapycal.core.workspace import Workspace
    
class NodeMeta(ABCMeta):
    class_def_counter = count()
    def_order = {}
    def __init__(self, name, bases, attrs):
        self.def_order[name] = next(self.class_def_counter)
        return super().__init__(name, bases, attrs)

class Node(SObject,metaclass=NodeMeta):
    frontend_type = 'Node'
    category = 'hidden'

    @classmethod
    def get_def_order(cls):
        return cls.def_order[cls.__name__]

    def build(self,is_preview=False,translation='0,0',**build_node_args):
        
        self.shape = self.add_attribute('shape', StringTopic, 'normal') # normal, simple, round
        self.output = self.add_attribute('output', ListTopic, [], is_stateful=False)
        self.label = self.add_attribute('label', StringTopic, 'Node', is_stateful=False)
        self.label_offset = self.add_attribute('label_offset', FloatTopic, 0)
        self.translation = self.add_attribute('translation', StringTopic,translation)
        self.is_preview = self.add_attribute('is_preview', IntTopic, 1 if is_preview else 0)
        self.category_ = self.add_attribute('category', StringTopic, self.category)
        self.exposed_attributes = self.add_attribute('exposed_attributes', ListTopic, [])
        self.running = self.add_attribute('running',IntTopic,1,is_stateful=False) # 0 for running, other for not running
        self.css_classes = self.add_attribute('css_classes',SetTopic,[])

        # for inspector
        self.type_topic = self.add_attribute('type', StringTopic, self.get_type_name())

        self.in_ports = self.add_attribute('in_ports', ObjListTopic[InputPort])
        self.out_ports = self.add_attribute('out_ports', ObjListTopic[OutputPort])

        self.controls = self.add_attribute('controls', ObjDictTopic[Control])

        '''
        Let user override build_node method instead of build method so that they don't have to call super().build(args) in their build method.
        '''
        self.build_node(**build_node_args)

    def build_node(self):
        '''
        Create attributes, ports, and controls here.
        
        Note: 
            This method will not be called when the object is being restored. The child objects will be restored automatically instead of
        running this method again.
        '''

    def init(self):
        '''
        This method is called after the node is built and its ports and controls are created. Use this method if you want to do something after
        the node is built.
        '''

        self.workspace:Workspace = self._server.globals.workspace
        self.old_node_info :NodeInfo|None = None
        
        from grapycal.sobjects.editor import Editor # import here to avoid circular import
        parent = self.get_parent()
        if isinstance(parent, Editor):
            self.editor = parent
        else:
            self.editor = None

        self.on('double_click', self.double_click, is_stateful=False)
        self.on('spawn', self.spawn , is_stateful=False)

        
        self._output_stream = OutputStream(self.raw_print)
        self._output_stream.set_event_loop(self.workspace.get_communication_event_loop())
        self.workspace.get_communication_event_loop().create_task(self._output_stream.run())

        self.init_node()

    def init_node(self):
        pass

    def restore_from_version(self,version:str,old:NodeInfo):
        '''
        Called when the node is created as a result of a old node being upgraded.
        The old node's information (including attribute values) is in the `old` argument.
        '''
        self.restore_attributes('translation')

    def restore_attributes(self,*attribute_names:str|tuple[str,str]):
        '''
        Recover attributes from the old node.
        '''
        assert self.old_node_info is not None
        for name in attribute_names:
            if isinstance(name,tuple):
                old_name,new_name = name
            else:
                old_name,new_name = name,name
            if not self.has_attribute(new_name):
                logger.warning(f'Attribute {new_name} does not exist in {self}')
                continue
            if not self.old_node_info.has_attribute(old_name):
                logger.warning(f'Attribute {old_name} does not exist in the old node of {self}')
                continue
            new_attr = self.get_attribute(new_name)
            old_attr = self.old_node_info[old_name]
            if isinstance(new_attr,WrappedTopic):
                new_attr.set_raw(old_attr)
            else:
                new_attr.set(old_attr)

    def restore_controls(self,*control_names:str|tuple[str,str]):
        '''
        Recover controls from the old node.
        '''
        assert self.old_node_info is not None
        for name in control_names:
            if isinstance(name,tuple):
                old_name,new_name = name
            else:
                old_name,new_name = name,name
            if not (new_name in self.controls):
                logger.warning(f'Control {new_name} does not exist in {self}')
                continue
            if not (old_name in self.old_node_info.controls):
                logger.warning(f'Control {old_name} does not exist in the old node of {self}')
                continue
            self.controls[new_name].restore_from(self.old_node_info.controls[old_name])

    def spawn(self, client_id):
        '''
        Called when a client wants to spawn a node.
        '''
        new_node = self.workspace.get_workspace_object().main_editor.create_node(type(self))
        new_node.add_tag(f'spawned_by_{client_id}') # So the client can find the node it spawned and make it follow the mouse

    def destroy(self) -> SObjectSerialized:
        '''
        Called when the node is destroyed. You can override this method to do something before the node is destroyed.
        Note: Overrided methods should call return super().destroy() at the end.
        '''
        self._output_stream.close()
        # remove all edges connected to the ports
        for port in self.in_ports:
            for edge in port.edges[:]:
                edge.remove()
        for port in self.out_ports:
            for edge in port.edges[:]:
                edge.remove()
        return super().destroy()

    def add_in_port(self,name:str,max_edges=64,display_name=None):
        '''
        Add an input port to the node.
        '''
        port = self.add_child(InputPort,name=name,max_edges=max_edges,display_name=display_name)
        self.in_ports.insert(port)
        return port

    def add_out_port(self,name:str,max_edges=64,display_name=None):
        '''
        Add an output port to the node.
        '''
        port = self.add_child(OutputPort,name=name,max_edges=max_edges,display_name=display_name)
        self.out_ports.insert(port)
        return port
    
    def remove_in_port(self,name:str):
        '''
        Remove an input port from the node.
        '''
        #find the port with the given name
        for port in self.in_ports:
            assert port is not None
            if port.name.get() == name:
                break
        else:
            raise ValueError(f'Port with name {name} does not exist')
        
        #remove all edges connected to the port
        for edge in port.edges[:]: 
            edge.remove() # do this in port.remove()?   

        #remove the port
        self.in_ports.remove(port)
        port.remove()

    def remove_out_port(self,name:str):
        '''
        Remove an output port from the node.
        '''
        #find the port with the given name
        for port in self.out_ports:
            assert port is not None
            if port.name.get() == name:
                break
        else:
            raise ValueError(f'Port with name {name} does not exist')
        
        #remove all edges connected to the port
        for edge in port.edges[:]: 
            edge.remove() # do this in port.remove()?

        #remove the port
        self.out_ports.remove(port)
        port.remove()

    def get_in_port(self,name:str) -> InputPort:
        '''
        Get an input port by its name.
        '''
        for port in self.in_ports:
            assert port is not None
            if port.name.get() == name:
                return port
        raise ValueError(f'Port with name {name} does not exist')
    
    def get_out_port(self,name:str) -> OutputPort:
        '''
        an output port by its name.
        '''
        for port in self.out_ports:
            assert port is not None
            if port.name.get() == name:
                return port
        raise ValueError(f'Port with name {name} does not exist')
    
    def has_in_port(self,name:str) -> bool:
        '''
        Check if an input port exists.
        '''
        for port in self.in_ports:
            assert port is not None
            if port.name.get() == name:
                return True
        return False
    
    def has_out_port(self,name:str) -> bool:
        '''
        Check if an output port exists.
        '''
        for port in self.out_ports:
            assert port is not None
            if port.name.get() == name:
                return True
        return False
    
    T = TypeVar('T', bound=Control)
    def add_control(self,control_type:type[T],name:str|None=None,**kwargs) -> T:
        '''
        Add a control to the node.
        '''
        if name is not None:
            if name in self.controls:
                raise ValueError(f'Control with name {name} already exists')
        else:
            name = 'Control0'
            i=0
            while name in self.controls:
                i+=1
                name = f'Control{i}'

        control = self.add_child(control_type,**kwargs)
        self.controls.add(name,control)
        return control
    
    def add_text_control(self,text:str='', label:str='',readonly=False, editable:bool=True,name:str|None=None) -> TextControl:
        '''
        Add a text control to the node.
        '''
        control = self.add_control(TextControl,text=text,label=label,readonly=readonly,editable=editable,name=name)
        return control
    
    def add_button_control(self,label:str='',name:str|None=None) -> ButtonControl:
        '''
        Add a button control to the node.
        '''
        control = self.add_control(ButtonControl,label=label,name=name)
        return control
    
    def add_image_control(self,name:str|None=None) -> ImageControl:
        '''
        Add an image control to the node.
        '''
        control = self.add_control(ImageControl,name=name)
        return control

    def remove_control(self,control:str|Control):
        if isinstance(control,str):
            control = self.controls[control]
        self.controls.remove(control)
        control.remove()
    
    # Wrap the SObject.addattribute() to make shorthand of exposing attributes after adding them.
    T1 = TypeVar("T1", bound=Topic|WrappedTopic)
    def add_attribute(
        self, topic_name:str, topic_type: type[T1], init_value=None, is_stateful=True,
        editor_type:str|None=None, display_name:str|None=None, **editor_args
        ) -> T1: 
        
        attribute = super().add_attribute(topic_name, topic_type, init_value, is_stateful)
        if editor_type is not None:
            self.expose_attribute(attribute,editor_type,display_name,**editor_args)
        return attribute
    
    def expose_attribute(self,attribute:Topic|WrappedTopic,editor_type,display_name=None,**editor_args):
        '''
        Expose an attribute to the editor.
        Args:
            - attribute: The attribute to expose.

            - editor_type: The type of the editor to use. Can be ``text`` or ``list``.

        '''
        if editor_args is None:
            editor_args = {}
        name = attribute.get_name()
        if display_name is None:
            display_name = name.split('/')[-1]
        editor_args['type'] = editor_type
        self.exposed_attributes.insert({
            'name':name,
            'display_name':display_name,
            'editor_args':editor_args
        })

    def print(self,*args,**kwargs):
        '''
        Print to the node's output.
        '''
        # print(*args,**kwargs,file=self._output_stream)
        # self._output_stream.flush()

        # maybe the self._output_stream can be abandoned
        output = io.StringIO()
        print(*args, file=output, **kwargs)
        contents = output.getvalue()
        output.close()

        self.raw_print(contents)

    def raw_print(self,data):
        if data=='':
            return
        if self.is_destroyed():
            logger.debug(f'Output received from a destroyed node {self.get_id()}: {data}')
        else:
            if len(self.output) > 100:
                self.output.set([])
                self.output.insert(['error','Too many output lines. Cleared.'])
            self.output.insert(['output',data])


    '''
    Run tasks in the background or foreground, redirecting stdout to the node's output stream.
    '''

    @contextmanager
    def _redirect_output(self):
        '''
        Returns a context manager that redirects stdout to the node's output stream.
        '''

        try:
            self._output_stream.enable_flush()
            with self.workspace.redirect(self._output_stream):
                yield
        finally:
            self._output_stream.disable_flush()

    def _run_in_background(self,task:Callable[[],None],to_queue=True,redirect_output=False):
        '''
        Run a task in the background thread.
        '''
        def task_wrapper():
            self.running.set(0)
            self.workspace.background_runner.set_exception_callback(lambda e:self._on_exception(e,truncate=3))
            if redirect_output:
                with self._redirect_output():
                    task()
            else:
                task()
            self.running.set(random.randint(0,10000))

        self.workspace.background_runner.push(task_wrapper,to_queue=to_queue)
        
    def _run_directly(self,task:Callable[[],None],redirect_output=False):
        '''
        Run a task in the current thread.
        '''
        self.running.set(0)
        try:
            if redirect_output:
                with self._redirect_output():
                    task()
            else:
                task()
        except Exception as e:
            self._on_exception(e,truncate=1)
        self.running.set(random.randint(0,10000))

    def run(self,task:Callable,background=True,to_queue=True,redirect_output=False,*args,**kwargs):
        '''
        Run a task in the node's context i.e. the stdout and errors will be redirected to the node's output attribute and be displayed in front-end.

        Args:
            - task: The task to run.

            - background: If set to True, the task will be scheduled to run in the background thread. Otherwise, it will be run in the current thread immediately.
            
            - to_queue: This argument is used only when `background` is True. If set to True, the task will be pushed to the :class:`.BackgroundRunner`'s queue.\
            If set to False, the task will be pushed to its stack. See :class:`.BackgroundRunner` for more details.
        '''
        task = functools.partial(task,*args,**kwargs)
        if background:
            self._run_in_background(task,to_queue,redirect_output=False)
        else:
            self._run_directly(task,redirect_output=False)

    def _on_exception(self, e, truncate=0):
        self.running.set(random.randint(0,10000))
        message = ''.join(traceback.format_exception(e)[truncate:])
        if self.is_destroyed():
            logger.warning(f'Exception occured in a destroyed node {self.get_id()}: {message}')
        else:
            if len(self.output) > 100:
                self.output.set([])
                self.output.insert(['error','Too many output lines. Cleared.'])
            self.output.insert(['error',message])


    def flash_running_indicator(self):
        self.running.set(0)
        self.running.set(random.randint(0,10000))

    '''
    Node events
    '''
    
    def edge_activated(self, edge:Edge, port:InputPort):
        '''
        Called when an edge on an input port is activated.
        '''
        pass

    def input_edge_added(self, edge:Edge, port:InputPort):
        '''
        Called when an edge is added to an input port.
        '''
        pass

    def input_edge_removed(self, edge:Edge, port:InputPort):
        '''
        Called when an edge is removed from an input port.
        '''
        pass

    def output_edge_added(self, edge:Edge, port:OutputPort):
        '''
        Called when an edge is added to an output port.
        '''
        pass

    def output_edge_removed(self, edge:Edge, port:OutputPort):
        '''
        Called when an edge is removed from an output port.
        '''
        pass

    def double_click(self):
        '''
        Called when the node is double clicked by an user.
        '''
        pass


from typing import Any, Dict
from grapycal.sobjects.controls.control import Control
from objectsync import StringTopic, EventTopic


class ButtonControl(Control):
    '''
    To add a button control to a node, use the following code in the node:
    ```python
    self.add_control(Button, label='')
    ```
    '''
    frontend_type = 'ButtonControl'
    def build(self, label:str=''):
        self.label = self.add_attribute('label', StringTopic, label)
        self._click = self.add_attribute('click', EventTopic, is_stateful=False)
        
    def init(self):
        self.on_click = self._click.on_emit
        


# python -m grapycal

from grapycal import GrapycalApp
import usersettings
import argparse

def main(s: usersettings.Settings):
    """
    Entry function of backend server
    """
    s.load_settings()
    app = GrapycalApp(s)
    app.run()

if __name__ == '__main__':
    #parse arguments
    parser = argparse.ArgumentParser(description='Grapycal backend server')
    parser.add_argument('--port', type=int, help='port to listen on')
    parser.add_argument('--host', type=str, help='host to listen on')
    parser.add_argument('--path', type=str, help='path to workspace file')
    parser.add_argument('--no-serve-webpage', action='store_true', help='if set, the server does not serve the webpage')
    parser.add_argument('--restart', action='store_true', help='if set, the workspace restarts when it exits. Convenient for development')
    args = parser.parse_args()
    s = usersettings.Settings("Grapycal")
    s.add_setting("port", int, default=8765) #type: ignore
    s.add_setting("host", str, default="localhost") #type: ignore
    s.add_setting("path", str, default="workspace.grapycal") #type: ignore
    s.load_settings()
    if args.port:
        s['port'] = args.port
    if args.host:
        s['host'] = args.host
    if args.path:
        s['path'] = args.path
    s.save_settings()
    s['no_serve_webpage'] = args.no_serve_webpage
    s['restart'] = args.restart
    main(s)
